@@ -1,13 +1,13 @@
 ##---------------------------------------------------------------------------##
 ##  File:
-##	@(#) mhamain.pl 2.22 01/04/10 21:36:49
+##	$Id: mhamain.pl,v 2.71 2003/08/13 03:03:17 ehood Exp $
 ##  Author:
-##      Earl Hood       mhonarc@pobox.com
+##      Earl Hood       mhonarc@mhonarc.org
 ##  Description:
 ##	Main library for MHonArc.
 ##---------------------------------------------------------------------------##
 ##    MHonArc -- Internet mail-to-HTML converter
-##    Copyright (C) 1995-2001	Earl Hood, mhonarc@pobox.com
+##    Copyright (C) 1995-2003	Earl Hood, mhonarc@mhonarc.org
 ##
 ##    This program is free software; you can redistribute it and/or modify
 ##    it under the terms of the GNU General Public License as published by
@@ -27,15 +27,38 @@
 
 package mhonarc;
 
-$VERSION = "2.4.8";
+require 5;
+
+$VERSION = '2.6.8';
 $VINFO =<<EndOfInfo;
-  MHonArc v$VERSION (Perl $])
-  Copyright (C) 1995-2001  Earl Hood, mhonarc\@pobox.com
+  MHonArc v$VERSION (Perl $] $^O)
+  Copyright (C) 1995-2003  Earl Hood, mhonarc\@mhonarc.org
   MHonArc comes with ABSOLUTELY NO WARRANTY and MHonArc may be copied only
   under the terms of the GNU General Public License, which may be found in
   the MHonArc distribution.
 EndOfInfo
 
+###############################################################################
+BEGIN {
+    ## Check what system we are executing under
+    require 'osinit.pl';  &OSinit();
+
+    ## Check if running setuid/setgid
+    $TaintMode = 0;
+    if ($UNIX && (( $< != $> ) || ( $( != $) ))) {
+	## We do not support setuid since there are too many
+	## security problems to handle, and if we did, mhonarc
+	## would probably not be very useful.
+	die "ERROR: setuid/setgid execution not supported!\n";
+
+	#$TaintMode = 1;
+	#$ENV{'PATH'}  = '/bin:/usr/bin';
+	#$ENV{'SHELL'} = '/bin/sh'  if exists $ENV{'SHELL'};
+	#delete @ENV{qw(IFS CDPATH ENV BASH_ENV)};
+    }
+}
+###############################################################################
+
 $CODE		= 0;
 $ERROR  	= "";
 @OrgARGV	= ();
@@ -49,7 +72,6 @@ my @_term_sigs	= qw(
     TERM TRAP USR1 USR2 VTALRM XCPU XFSZ
 );
 
-
 ###############################################################################
 ##	Public routines
 ###############################################################################
@@ -62,9 +84,6 @@ sub initialize {
     ##	Turn off buffered I/O to terminal
     my($curfh) = select(STDOUT);  $| = 1;  select($curfh);
 
-    ##	Check what system we are executing under
-    require 'osinit.pl';  &OSinit();
-
     ##	Require essential libraries
     require 'mhlock.pl';
     require 'mhopt.pl';
@@ -230,7 +249,7 @@ sub doit {
     ## HTML message listing to standard output.
     if ($IDXONLY) {
 	IDXPAGE: {
-	    &compute_page_total();
+	    compute_page_total();
 	    if ($IdxPageNum && $MULTIIDX) {
 		if ($IdxPageNum =~ /first/i) {
 		    $IdxPageNum = 1;
@@ -248,20 +267,18 @@ sub doit {
 	    $NumOfPages = 1;
 	}
 	if ($THREAD) {
-	    &compute_threads();
-	    &write_thread_index($IdxPageNum);
+	    compute_threads();
+	    write_thread_index($IdxPageNum);
 	} else {
-	    &write_main_index($IdxPageNum);
+	    write_main_index($IdxPageNum);
 	}
 	return 1;
     }
 
     ## Get here, we are processing mail folders
+    my($index, $fields, $fh, $cur_msg_cnt);
 
-    local($mesg, $tmp, $index, $sub, $from, $i, $date, $fh);
-    local(%fields);
-
-    $i = $NumOfMsgs;
+    $cur_msg_cnt = $NumOfMsgs;
     ##-------------------##
     ## Read mail folders ##
     ##-------------------##
@@ -275,34 +292,20 @@ sub doit {
 	$handle = $ADD;
 
 	## Read mail head
-	($index,$from,$date,$sub,$header) =
-	    &read_mail_header($handle, *mesg, *fields);
-
-	if ($index ne '') {
-	    ($From{$index},$Date{$index},$Subject{$index}) =
-		($from,$date,$sub);
+	($index, $fields) = read_mail_header($handle);
 
+	if ($index) {
 	    $AddIndex{$index} = 1;
-	    $IndexNum{$index} = &getNewMsgNum();
-
-	    $MsgHead{$index} = $mesg;
-
-	    ## Read rest of message
-	    $Message{$index} = &read_mail_body(
-					$handle,
-					$index,
-					$header,
-				        *fields,
-					$NoMsgPgs);
+	    read_mail_body($handle, $index, $fields, $NoMsgPgs);
 	}
 
     ## Adding/converting mail{boxes,folders}
     } else {
 	print STDOUT ($ADD ? "Adding" : "Converting"), " messages to $OUTDIR"
 	    unless $QUIET;
-	local($mbox, $mesgfile, @files);
+	my($mbox, $mesgfile, @files);
 
-	foreach $mbox (@ARGV) {
+	MAILFOLDER: foreach $mbox (@ARGV) {
 
 	    ## MH mail folder (a directory)
 	    if (-d $mbox) {
@@ -315,35 +318,29 @@ sub doit {
 		@files = sort { $a <=> $b } grep(/$MHPATTERN/o,
 						 readdir(MAILDIR));
 		closedir(MAILDIR);
-		foreach (@files) {
+
+		local($_);
+		MHFILE: foreach (@files) {
 		    $mesgfile = "${mbox}${DIRSEP}${_}";
-		    if (!($fh = &file_open($mesgfile))) {
-			warn "\nWarning: Unable to open message $mesgfile\n";
-			next;
+		    eval {
+			$fh = file_open($mesgfile);
+		    };
+		    if ($@) {
+			warn $@,
+			     qq/...Skipping "$mesgfile"\n/;
+			next MHFILE;
 		    }
 		    print STDOUT "."  unless $QUIET;
-		    $mesg = '';
-		    ($index,$from,$date,$sub,$header) =
-			&read_mail_header($fh, *mesg, *fields);
+		    ($index, $fields) = read_mail_header($fh);
 
 		    #  Process message if valid
-		    if ($index ne '') {
-			($From{$index},$Date{$index},$Subject{$index}) =
-			    ($from,$date,$sub);
-			$MsgHead{$index} = $mesg;
-
+		    if ($index) {
 			if ($ADD && !$SLOW) { $AddIndex{$index} = 1; }
-			$IndexNum{$index} = &getNewMsgNum();
+			read_mail_body($fh, $index, $fields, $NoMsgPgs);
 
-			$Message{$index} = &read_mail_body(
-						$fh,
-						$index,
-						$header,
-					        *fields,
-						$NoMsgPgs);
 			#  Check if conserving memory
 			if ($SLOW && $DoArchive) {
-			    &output_mail($index, 1, 1);
+			    output_mail($index, 1, 1);
 			    $Update{$IndexNum{$index}} = 1;
 			}
 			if ($SLOW || !$DoArchive) {
@@ -358,9 +355,15 @@ sub doit {
 	    } else {
 		if ($mbox eq "-") {
 		    $fh = $MhaStdin;
-		} elsif (!($fh = &file_open($mbox))) {
-		    warn "\nWarning: Unable to open $mbox\n";
-		    next;
+		} else {
+		    eval {
+			$fh = file_open($mbox);
+		    };
+		    if ($@) {
+			warn $@,
+			     qq/...Skipping "$mbox"\n/;
+			next MAILFOLDER;
+		    }
 		}
 
 		$MBOX = 1;  $MH = 0;
@@ -368,26 +371,14 @@ sub doit {
 		# while (<$fh>) { last if /$FROM/o; }
 		MBOX: while (!eof($fh)) {
 		    print STDOUT "."  unless $QUIET;
-		    $mesg = '';
-		    ($index,$from,$date,$sub,$header) =
-			&read_mail_header($fh, *mesg, *fields);
-
-		    if ($index ne '') {
-			($From{$index},$Date{$index},$Subject{$index}) =
-			    ($from,$date,$sub);
-			$MsgHead{$index} = $mesg;
+		    ($index, $fields) = read_mail_header($fh);
 
+		    if ($index) {
 			if ($ADD && !$SLOW) { $AddIndex{$index} = 1; }
-			$IndexNum{$index} = &getNewMsgNum();
+			read_mail_body($fh, $index, $fields, $NoMsgPgs);
 
-			$Message{$index} = &read_mail_body(
-						$fh,
-						$index,
-						$header,
-						*fields,
-						$NoMsgPgs);
 			if ($SLOW && $DoArchive) {
-			    &output_mail($index, 1, 1);
+			    output_mail($index, 1, 1);
 			    $Update{$IndexNum{$index}} = 1;
 			}
 			if ($SLOW || !$DoArchive) {
@@ -396,7 +387,7 @@ sub doit {
 			}
 
 		    } else {
-			&read_mail_body($fh, $index, $header, *fields, 1);
+			read_mail_body($fh, $index, $fields, 1);
 		    }
 		}
 		close($fh);
@@ -412,10 +403,12 @@ sub doit {
     }
 
     ## Check if there are any new messages
-    if (!$EDITIDX && ($i == $NumOfMsgs)) {
+    if (!$EDITIDX && ($cur_msg_cnt > 0) &&
+	    !scalar(%AddIndex) && !scalar(%Update)) {
 	print STDOUT "No new messages\n"  unless $QUIET;
 	return 1;
     }
+    $NewMsgCnt = $NumOfMsgs - $cur_msg_cnt;
 
     ## Write pages
     &write_pages();
@@ -426,8 +419,9 @@ sub doit {
 ##	write_pages writes out all archive pages and db
 ##
 sub write_pages {
-    local($i, $key, $index, $tmp, $tmp2);
+    my($i, $j, $key, $index, $tmp, $tmp2);
     my(@array2);
+    my($mloc, $tloc);
 
     ## Remove old message if hit maximum size or expiration
     if (($MAXSIZE && ($NumOfMsgs > $MAXSIZE)) ||
@@ -445,7 +439,6 @@ sub write_pages {
 	&ign_signals();
 
 	## Expiration based upon time
-	my($mloc, $tloc);
 	foreach $index (sort_messages(0,0,0,0)) {
 	    last  unless
 		    ($MAXSIZE && ($NumOfMsgs > $MAXSIZE)) ||
@@ -472,7 +465,7 @@ sub write_pages {
 		    $Update{$IndexNum{$TListOrder[$tloc+$i]}} = 1
 			if $tloc-$i >= $#TListOrder;
 		}
-		foreach (split(/$bs/o, $FollowOld{$index})) {
+		foreach (@{$FollowOld{$index}}) {
 		    $Update{$IndexNum{$_}} = 1;
 		}
 	    }
@@ -494,10 +487,10 @@ sub write_pages {
     @Index2MLoc{@MListOrder} = (0 .. $#MListOrder);
 
     ## Compute follow up messages
-    &compute_follow_ups(\@MListOrder);
+    compute_follow_ups(\@MListOrder);
 
-    ##	Compute thread information (sets ThreadList, TListOrder, Index2TLoc)
-    &compute_threads();
+    ## Compute thread information (sets ThreadList, TListOrder, Index2TLoc)
+    compute_threads();
 
     ## Check for which messages to update when adding to archive
     if ($ADD) {
@@ -510,7 +503,7 @@ sub write_pages {
 	    $i = 0;
 	    foreach $index (@MListOrder) {
 		## Check for New follow-up links
-		if ($FollowOld{$index} ne $Follow{$index}) {
+		if (is_follow_ups_diff($index)) {
 		    $Update{$IndexNum{$index}} = 1;
 		}
 		## Check if new message; must update links in prev/next msgs
@@ -530,7 +523,7 @@ sub write_pages {
 			if $i < $#MListOrder;
 		}
 		## Check for New reference links
-		foreach (split(/$X/o, $Refs{$index})) {
+		foreach (@{$Refs{$index}}) {
 		    $tmp = $MsgId{$_};
 		    if (defined($IndexNum{$tmp}) && $AddIndex{$tmp}) {
 			$Update{$IndexNum{$index}} = 1;
@@ -555,6 +548,16 @@ sub write_pages {
 			if $i > 0;
 		    $Update{$IndexNum{$TListOrder[$i+1]}} = 1
 			if $i < $#TListOrder;
+
+		    $tloc = $Index2TLoc{$index};
+		    for ($j=2; $j <= $TSliceNBefore; ++$j) {
+			$Update{$IndexNum{$TListOrder[$tloc-$j]}} = 1
+			    if $tloc-$j >= 0;
+		    }
+		    for ($j=2; $j <= $TSliceNAfter; ++$j) {
+			$Update{$IndexNum{$TListOrder[$tloc+$j]}} = 1
+			    if $tloc-$j >= $#TListOrder;
+		    }
 		}
 		$i++;
 	    }
@@ -563,7 +566,7 @@ sub write_pages {
 
     ##	Compute total number of pages
     $i = $NumOfPages;
-    &compute_page_total();
+    compute_page_total();
 
     ## Update all pages for $LASTPG$
     if ($UsingLASTPG && ($i != $NumOfPages)) {
@@ -574,17 +577,17 @@ sub write_pages {
     ##------------##
     ## Write Data ##
     ##------------##
-    &ign_signals();		# Ignore termination signals
+    ign_signals();		# Ignore termination signals
     print STDOUT "\n"  unless $QUIET;
 
     ## Write indexes and mail
-    &write_mail()		unless $NoMsgPgs;
-    &write_main_index() 	if $MAIN;
-    &write_thread_index()	if $THREAD;
+    write_mail()		unless $NoMsgPgs;
+    write_main_index()  	if $MAIN;
+    write_thread_index()	if $THREAD;
 
     ## Write database
     print STDOUT "Writing database ...\n"  unless $QUIET;
-    &output_db($DBPathName);
+    output_db($DBPathName);
 
     ## Write any alternate indexes
     $IdxMinPg = 0; $TIdxMinPg = 0;
@@ -594,41 +597,56 @@ sub write_pages {
 
 	## find other index resource file
 	IDXFIND: {
-	    if (! -e $rc) {
-		$rcfile = join($DIRSEP, $OUTDIR, $rc);
-		if (! -e $rcfile) {
-		    # look thru @INC to find file
-		    local($_);
-		    foreach (@INC) {
-			$rcfile = join($DIRSEP, $_, $rc);
-			if (-e $rcfile) {
-			    last IDXFIND;
-			}
-		    }
-		    warn qq/Warning: Unable to find resource file "$rc"\n/;
-		    next OTHERIDX;
-		}
-
-	    } else {
+	    if (-e $rc) {
+		# in current working directory
 		$rcfile = $rc;
+		last IDXFIND;
 	    }
+	    if (defined $MainRcDir) {
+		# check if located with main resource file
+		$rcfile = join($DIRSEP, $MainRcDir, $rc);
+		last IDXFIND  if -e $rcfile;
+	    }
+	    if (defined $ENV{'HOME'}) {
+		# check if in home directory
+		$rcfile = join($DIRSEP, $ENV{'HOME'}, $rc);
+		last IDXFIND  if -e $rcfile;
+	    }
+
+	    # check if in archive directory
+	    $rcfile = join($DIRSEP, $OUTDIR, $rc);
+	    last IDXFIND  if -e $rcfile;
+
+	    # look thru @INC to find file
+	    local($_);
+	    foreach (@INC) {
+		$rcfile = join($DIRSEP, $_, $rc);
+		if (-e $rcfile) {
+		    last IDXFIND;
+		}
+	    }
+	    warn qq/Warning: Unable to find resource file "$rc"\n/;
+	    next OTHERIDX;
 	}
 	    
 	## read resource file and print index
-	if (&read_fmt_file($rcfile)) {
+	if (read_resource_file($rcfile)) {
 	    if ($THREAD) {
 		@TListOrder = ();
-		&write_thread_index();
+		write_thread_index();
 	    } else {
 		@MListOrder = ();
-		&write_main_index();
+		write_main_index();
 	    }
 	}
     }
 
-    print STDOUT "$NumOfMsgs messages\n"  unless $QUIET;
+    unless ($QUIET) {
+	print STDOUT "$NewMsgCnt new messages\n"  if $NewMsgCnt > 0;
+	print STDOUT "$NumOfMsgs total messages\n";
+    }
 
-}
+} ## End: write_pages()
 
 ##---------------------------------------------------------------------------
 ##	Compute follow-ups
@@ -636,20 +654,17 @@ sub write_pages {
 sub compute_follow_ups {
     my $idxlst = shift;
     my($index, $tmp, $tmp2);
-    my(@array);
 
     %Follow = ();
     foreach $index (@$idxlst) {
 	$FolCnt{$index} = 0  unless $FolCnt{$index};
-	if ( defined($Refs{$index}) &&
-	     (@array = split(/$X/o, $Refs{$index}))) {
-
-	    $tmp2 = $array[$#array];
+	if (defined($Refs{$index}) && scalar(@{$Refs{$index}})) {
+	    $tmp2 = $Refs{$index}->[-1];
 	    next  unless defined($MsgId{$tmp2}) &&
 			 defined($IndexNum{$MsgId{$tmp2}});
 	    $tmp = $MsgId{$tmp2};
-	    if ($Follow{$tmp}) { $Follow{$tmp} .= $bs . $index; }
-	    else { $Follow{$tmp} = $index; }
+	    if ($Follow{$tmp}) { push(@{$Follow{$tmp}}, $index); }
+	    else { $Follow{$tmp} = [ $index ]; }
 	    ++$FolCnt{$tmp};
 	}
     }
@@ -673,7 +688,7 @@ sub compute_page_total {
 ##	array containing indexes of messages to output.
 ##
 sub write_mail {
-    local($hack) = (0);
+    my($hack) = (0);
     print STDOUT "Writing mail "  unless $QUIET;
 
     if ($SLOW && !$ADD) {
@@ -683,7 +698,7 @@ sub write_mail {
 
     foreach $index (@MListOrder) {
 	print STDOUT "."  unless $QUIET;
-	&output_mail($index, $AddIndex{$index}, 0);
+	output_mail($index, $AddIndex{$index}, 0);
     }
 
     if ($hack) {
@@ -695,40 +710,46 @@ sub write_mail {
 
 ##---------------------------------------------------------------------------
 ##	read_mail_header() is responsible for parsing the header of
-##	a mail message.
+##	a mail message and loading message information into hash
+##	structures.
+##
+##	($index, $header_fields_ref) = read_mail_header($filehandle);
 ##
 sub read_mail_header {
-    local($handle, *mesg, *fields) = @_;
-    my(%l2o, $header, $index, $date, $tmp, @array);
-    local(@refs);
-    local($from, $sub, $msgid);
+    my $handle = shift;
+    my($date, $tmp, $i, $field, $value);
+    my($from, $sub, $msgid, $ctype);
     local($_);
 
-    $header = &readmail::MAILread_file_header($handle, *fields, *l2o);
-    @refs = ();
-    @array = ();
+    my $index  = undef;
+    my $msgnum = undef;
+    my @refs   = ();
+    my @array  = ();
+    my($fields, $header) = readmail::MAILread_file_header($handle);
 
     ##---------------------------##
     ## Check for no archive flag ##
     ##---------------------------##
     if ( $CheckNoArchive &&
-	 ($fields{'restrict'} =~ /no-external-archive/i ||
-	  $fields{'x-no-archive'} =~ /yes/i) ) {
-	return ("", "", "", "", "");
+	 ((defined($fields->{'restrict'}) &&
+	  grep { /no-external-archive/i } @{$fields->{'restrict'}}) ||
+	  (defined($fields->{'x-no-archive'}) &&
+	   grep { /yes/i } @{$fields->{'x-no-archive'}})) ) {
+	return undef;
     }
 
     ##----------------------------------##
     ## Check for user-defined exclusion ##
     ##----------------------------------##
-    if ( $MsgExcFilter ) {
-	return ("", "", "", "", "") if &mhonarc::message_exclude($header);
+    if ($MsgExcFilter) {
+	return undef  if mhonarc::message_exclude($header);
     }
 
     ##------------##
     ## Get Msg-ID ##
     ##------------##
-    $msgid = $fields{'message-id'} || $fields{'msg-id'} || 
-	     $fields{'content-id'};
+    $msgid = $fields->{'message-id'}[0] || $fields->{'msg-id'}[0] || 
+	     $fields->{'content-id'}[0];
     if (defined($msgid)) {
 	if ($msgid =~ /<([^>]*)>/) {
 	    $msgid = $1;
@@ -753,35 +774,47 @@ sub read_mail_header {
     }
 
     ## Return if message already exists in archive
-    if ($msgid && defined($MsgId{$msgid})) {
-	return ("", "", "", "", "");
+    if ($msgid && defined($index = $MsgId{$msgid})) {
+	if ($Reconvert) {
+	    $msgnum = $IndexNum{$index};
+	    delmsg($index);
+	    $index = undef;
+	} else {
+	    return undef;
+	}
     }
 
     ##----------##
     ## Get date ##
     ##----------##
-    $date = "";  $index = "";
-    foreach (@DateFields) {
-	next  unless $fields{$_};
+    $date = "";
+    foreach (@_DateFields) {
+	($field, $i) = @{$_}[0,1];
+	next  unless defined($fields->{$field}) &&
+		     defined($value = $fields->{$field}[$i]);
 
 	## Treat received field specially
-	if ($_ eq 'received') {
-	    @array = split(/;/,
-			   (split(/$readmail::FieldSep/o, $fields{$_}))[0]);
+	if ($field eq 'received') {
+	    @array = split(/;/, $value);
+#	    if ((scalar(@array) <= 1) || (scalar(@array) > 2)) {
+#		warn qq/\nWarning: Received header field looks improper:\n/,
+#		       qq/         Received: $value\n/,
+#		       qq/         Message-Id: <$msgid>\n/;
+#	    }
 	    $date = pop @array;
 	## Any other field should just be a date
 	} else {
-	    $date = (split(/$readmail::FieldSep/o, $fields{$_}))[0];
+	    $date = $value;
 	}
 	$date =~ s/^\s+//;  $date =~ s/\s+$//;
 
 	## See if time_t can be determined.
-	if (($date =~ /\w/) && (@array = &parse_date($date))) {
-	    $index = &get_time_from_date(@array[1..$#array]);
+	if (($date =~ /\w/) && (@array = parse_date($date))) {
+	    $index = get_time_from_date(@array[1..$#array]);
 	    last;
 	}
     }
-    if ($index eq "") {
+    if (!$index) {
 	warn qq/\nWarning: Could not parse date for message\n/,
 	       qq/         Message-Id: <$msgid>\n/;
 	# Use current time
@@ -790,19 +823,19 @@ sub read_mail_header {
 	$date  = &time2str("", $index, 1)  unless $date =~ /\S/;
     }
 
-    ## Return if message too old to add
+    ## Return if message too old to add (note, $index just contains time).
     if (&expired_time($index)) {
-	return ("", "", "", "", "");
+	return undef;
     }
 
     ##-------------##
     ## Get Subject ##
     ##-------------##
-    if (defined($fields{'subject'}) and $fields{'subject'} =~ /\S/) {
-	($sub = $fields{'subject'}) =~ s/\s+$//;
+    if (defined($fields->{'subject'}) && ($fields->{'subject'}[0] =~ /\S/)) {
+	($sub = $fields->{'subject'}[0]) =~ s/\s+$//;
 	$sub = subject_strip($sub)  if $SubStripCode;
     } else {
-	$sub = 'No Subject';
+	$sub = '';
     }
 
     ##----------##
@@ -810,74 +843,110 @@ sub read_mail_header {
     ##----------##
     $from = "";
     foreach (@FromFields) {
-	next  unless $fields{$_};
-	$from = $fields{$_};
+	next  unless defined $fields->{$_};
+	$from = $fields->{$_}[0];
 	last;
     }
-    $from = 'Anonymous'  unless $from;
+    $from = 'Unknown'  unless $from;
 
     ##----------------##
     ## Get References ##
     ##----------------##
-    if (defined($tmp = $fields{'references'})) {
+    if (defined($fields->{'references'})) {
+	$tmp = $fields->{'references'}[0];
 	while ($tmp =~ s/<([^<>]+)>//) {
 	    push(@refs, $1);
 	}
     }
-    if (defined($tmp = $fields{'in-reply-to'})) {
-	my $irtoid = "";
-	while ($tmp =~ s/<([^<>]+)>//) { $irtoid = $1 };
-	push(@refs, $irtoid)  if $irtoid;
+    if (defined($fields->{'in-reply-to'})) {
+	my $irtoid;
+	foreach (@{$fields->{'in-reply-to'}}) {
+	    $tmp = $_;
+	    $irtoid = "";
+	    while ($tmp =~ s/<([^<>]+)>//) { $irtoid = $1 };
+	    push(@refs, $irtoid)  if $irtoid;
+	}
     }
+    @refs = remove_dups(\@refs);        # Remove duplicate msg-ids
 
-    ##------------------------##
-    ## Create HTML for header ##
-    ##------------------------##
-    $mesg .= &htmlize_header(*fields, *l2o);
+    ##------------------##
+    ## Get Content-Type ##
+    ##------------------##
+    if (defined($fields->{'content-type'})) {
+	($ctype = $fields->{'content-type'}[0]) =~ m%^\s*([\w\-\./]+)%;
+	$ctype = lc ($1 || 'text/plain');
+    } else {
+	$ctype = 'text/plain';
+    }
 
     ## Insure uniqueness of index
-    $index .= $X . sprintf("%d",$LastMsgNum+1);
+    $index .= $X . sprintf('%d',(defined($msgnum)?$msgnum:($LastMsgNum+1)));
 
-    if ($fields{'content-type'}) {
-	($tmp = $fields{'content-type'}) =~ m%^\s*([\w-\./]+)%;
-	$tmp = $1 || 'text/plain';
-	$tmp =~ tr/A-Z/a-z/;
-    } else {
-	$tmp = 'text/plain';
+    ## Set mhonarc fields.  Note how values are NOT arrays.
+    $fields->{'x-mha-index'} = $index;
+    $fields->{'x-mha-message-id'} = $msgid;
+    $fields->{'x-mha-from'} = $from;
+    $fields->{'x-mha-subject'} = $sub;
+    $fields->{'x-mha-content-type'} = $ctype;
+
+    ## Invoke callback if defined
+    if (defined($CBMessageHeadRead) && defined(&$CBMessageHeadRead)) {
+	return undef  unless &$CBMessageHeadRead($fields, $header);
     }
-    $ContentType{$index} = $tmp;
 
+    $From{$index} = $from;
+    $Date{$index} = $date;
+    $Subject{$index} = $sub;
+    $MsgHead{$index} = htmlize_header($fields);
+    $ContentType{$index} = $ctype;
     if ($msgid) {
 	$MsgId{$msgid} = $index;
 	$NewMsgId{$msgid} = $index;	# Track new message-ids
 	$Index2MsgId{$index} = $msgid;
     }
-    &remove_dups(*refs);                # Remove duplicate msg-ids
-    $Refs{$index} = join($X, @refs)  if (@refs);
+    if (defined($msgnum)) {
+	$IndexNum{$index} = $msgnum;
+	++$NumOfMsgs; # Counteract decrement by delmsg
+    } else {
+	$IndexNum{$index} = getNewMsgNum();
+    }
 
-    ($index,$from,$date,$sub,$header);
+    $Refs{$index} = [ @refs ]  if (@refs);
+
+    ## Grab any extra fields to store
+    foreach $field (@ExtraHFields) {
+	next  unless $fields->{$field};
+	if (!defined($tmp = $ExtraHFields{$index})) {
+	    $tmp = $ExtraHFields{$index} = { };
+	}
+	if ($HFieldsAddr{$field}) {
+	    $tmp->{$field} = join(', ', @{$fields->{$field}});
+	} else {
+	    $tmp->{$field} = join(' ', @{$fields->{$field}});
+	}
+    }
+
+    ($index, $fields);
 }
 
 ##---------------------------------------------------------------------------
 ##	read_mail_body() reads in the body of a message.  The returned
 ##	filtered body is in $ret.
 ##
+##	$html = read_mail_body($fh, $index, $fields_hash_ref,
+##			       $skipConversion);
+##
 sub read_mail_body {
-    local($handle, $index, $header, *fields, $skip) = @_;
+    my($handle, $index, $fields, $skip) = @_;
     my($ret, $data) = ('', '');
-    my(@files) = ();
+    my(@files);
     local($_);
 
-    ## Define "globals" for use by filters
-    ##	NOTE: This stuff can be handled better, and will be done
-    ##	      when/if I get around to rewriting mhonarc in (OO) Perl 5.
-    $MHAmsgnum = &fmt_msgnum($IndexNum{$index}) unless $skip;
-
     ## Slurp up message body
     ##	UUCP mailbox
     if ($MBOX) {
-	if ($CONLEN && $fields{"content-length"}) { # Check for content-length
-	    my($len, $cnt) = ($fields{"content-length"}, 0);
+	if ($CONLEN && defined($fields->{"content-length"})) {
+	    my($len, $cnt) = ($fields->{"content-length"}[0], 0);
 	    if ($len) {
 		while (<$handle>) {
 		    $cnt += length($_);		# Increment byte count
@@ -903,28 +972,45 @@ sub read_mail_body {
 	$data = <$handle>;
     }
 
-    ## Filter data
     return ''  if $skip;
-    $fields{'content-type'} = 'text/plain'
-	if (!defined($fields{'content-type'})) ||
-	   ($fields{'content-type'} !~ /\S/);
-    ($ret, @files) = &readmail::MAILread_body($header, $data,
-				    $fields{'content-type'},
-				    $fields{'content-transfer-encoding'});
-    if ($ret) {
-	# no-op
+
+    ## Invoke callback if defined
+    if (defined($CBRawMessageBodyRead) && defined(&$CBRawMessageBodyRead)) {
+	&$CBRawMessageBodyRead($fields, \$data);
+    }
+
+    ## Define "globals" for use by filters
+    ##	NOTE: This stuff can be handled better, and will be done
+    ##	      when/if I get around to rewriting mhonarc in (OO) Perl 5.
+    $MHAindex  = $index;
+    $MHAmsgnum = &fmt_msgnum($IndexNum{$index});
+    $MHAmsgid  = $Index2MsgId{$index};
+
+    ## Filter data
+    ($ret, @files) = &readmail::MAILread_body($fields, \$data);
+    $ret = ''     unless defined $ret;
+    @files = ( )  unless @files;
+
+    ## Invoke callback if defined
+    if (defined($CBMessageBodyRead) && defined(&$CBMessageBodyRead)) {
+	&$CBMessageBodyRead($fields, \$ret, \@files);
+	$Message{$index} = $ret;
     } else {
-	$ret = join('',
-		    "<DL>\n",
-		    "<DT><STRONG>Warning</STRONG></DT>\n",
-		    "<DD>Could not process message with given Content-Type: \n",
-		    "<CODE>", &htmlize($fields{'content-type'}), "</CODE>\n",
-		    "</DD>\n",
-		    "</DL>\n"
-		    );
+	$Message{$index} = $ret;
+    }
+
+    if (!defined($ret) || $ret eq '') {
+	warn qq/\n/,
+	     qq/Warning: Empty body data generated:\n/,
+	     qq/         Message-Id: $MHAmsgid\n/,
+	     qq/         Message Number: $MHAmsgnum\n/,
+	     qq/         Content-Type/,
+			 ($fields->{'content-type'}[0] || 'text/plain'),
+			 qq/\n/;
+	$ret = '';
     }
     if (@files) {
-	$Derived{$index} = join($X, @files);
+	$Derived{$index} = [ @files ];
     }
     $ret;
 }
@@ -935,28 +1021,50 @@ sub read_mail_body {
 ##	    $force	=> flag if mail is written and not editted, regardless
 ##	    $nocustom	=> ignore sections with user customization
 ##
+##	This function returns ($msgnum, $filename) if everything went
+##	okay, but no calls to this routine check the return values.
+##
 sub output_mail {
-    local($index, $force, $nocustom) = @_;
-    local($msgi, $tmp, $tmp2, $template, @array2);
-    local($filepathname, $tmppathname, $i_p0, $filename, $msghandle,
-	  $msginfh, $drvfh);
-    local($adding) = ($ADD && !$force && !$SINGLE);
+    my($index, $force, $nocustom) = @_;
+    my($msgi, $tmp, $tmp2, $template, @array2);
+    my($msghandle, $msginfh);
 
-    $i_p0 = &fmt_msgnum($IndexNum{$index});
+    my $msgnum	     = $IndexNum{$index};
+    if (!$SINGLE && !defined($msgnum)) {
+      # Something bad must have happened to message, so we just
+      # quietly return.
+      return;
+    }
 
-    $filename     = &msgnum_filename($IndexNum{$index});
-    $filepathname = join($DIRSEP, $OUTDIR, $filename);
-    $tmppathname  = join($DIRSEP, $OUTDIR, "msgtmp.$$");
+    my $adding	     = ($ADD && !$force && !$SINGLE);
+    my $i_p0 	     = fmt_msgnum($msgnum);
+    my $filename     = msgnum_filename($msgnum);
+    my $filepathname = join($DIRSEP, $OUTDIR, $filename);
+    my $tmppathname;
 
     if ($adding) {
-	return ($i_p0,$filename)  unless $Update{$IndexNum{$index}};
+	return ($i_p0, $filename)  unless $Update{$msgnum};
 	#&file_rename($filepathname, $tmppathname);
-	$msginfh = &file_open($filepathname);
+	eval {
+	  $msginfh = file_open($filepathname);
+	};
+	if ($@) {
+	  # Something is screwed up with archive: We try to delete
+	  # message from database since message file appears to have
+	  # disappeared
+	  warn $@,
+	       qq/...Will attempt to remove message and continue on\n/;
+	  delmsg($index);
+
+	  # Nothing else to do, so return.
+	  return;
+	}
     }
     if ($SINGLE) {
-	$msghandle = 'STDOUT';
+	$msghandle = \*STDOUT;
     } else {
-	$msghandle = &file_create($tmppathname, $GzipFiles);
+	($msghandle, $tmppathname) =
+	    file_temp('tmsg'.$i_p0.'_XXXXXXXXXX', $OUTDIR);
     }
 
     ## Output HTML header
@@ -968,8 +1076,11 @@ sub output_mail {
     if (!$nocustom) {
 	#&defineIndex2MsgId();
 
-	($template = $SSMARKUP) =~ s/$VarExp/&replace_li_var($1,$index)/geo;
-	print $msghandle $template;
+	$template = ($MSGPGSSMARKUP ne '') ? $MSGPGSSMARKUP : $SSMARKUP;
+	if ($template ne '') {
+	    $template =~ s/$VarExp/&replace_li_var($1,$index)/geo;
+	    print $msghandle $template;
+	}
 
 	# Output comments -- more informative, but can be used for
 	#		     error recovering.
@@ -983,16 +1094,15 @@ sub output_mail {
 		  #ContentType
 
 	if (defined($Refs{$index})) {
-	    foreach (split(/$X/o, $Refs{$index})) {
+	    foreach (@{$Refs{$index}}) {
 		print $msghandle
-		    "<!--X-Reference: ", &commentize($_), " -->\n";
+		    "<!--X-Reference: ", commentize($_), " -->\n";
 			  #Reference-Id
 	    }
 	}
 	if (defined($Derived{$index})) {
-	    foreach (split(/$X/o, $Derived{$index})) {
-		print $msghandle
-		    "<!--X-Derived: ", &commentize($_), " -->\n";
+	    foreach (@{$Derived{$index}}) {
+		print $msghandle "<!--X-Derived: ", commentize($_), " -->\n";
 	    }
 	}
 	print $msghandle "<!--X-Head-End-->\n";
@@ -1033,7 +1143,7 @@ sub output_mail {
 	while (<$msginfh>) {
 	    # check if subject header delimited
 	    if (/<!--X-Subject-Header-Begin/) {
-		$tmp2 =~ s%($AddrExp)%&link_refmsgid($1,1)%geo;
+		$tmp2 =~ s/($HAddrExp)/&link_refmsgid($1,1)/geo;
 		print $msghandle $tmp2;
 		$tmp2 = "";
 
@@ -1049,7 +1159,7 @@ sub output_mail {
 	    }
 	    # check if head/body separator delimited
 	    if (/<!--X-Head-Body-Sep-Begin/) {
-		$tmp2 =~ s%($AddrExp)%&link_refmsgid($1,1)%geo;
+		$tmp2 =~ s/($HAddrExp)/&link_refmsgid($1,1)/geo;
 		print $msghandle $tmp2;
 		$tmp2 = "";
 
@@ -1067,7 +1177,7 @@ sub output_mail {
 	    $tmp2 .= $_;
 	    last  if /<!--X-MsgBody-End/;
 	}
-	$tmp2 =~ s%($AddrExp)%&link_refmsgid($1,1)%geo;
+	$tmp2 =~ s/($HAddrExp)/&link_refmsgid($1,1)/geo;
 	print $msghandle $tmp2;
 
     } else {
@@ -1078,8 +1188,8 @@ sub output_mail {
 	print $msghandle $template;
 	print $msghandle "<!--X-Subject-Header-End-->\n";
 
-	$MsgHead{$index} =~ s%($AddrExp)%&link_refmsgid($1)%geo;
-	$Message{$index} =~ s%($AddrExp)%&link_refmsgid($1)%geo;
+	$MsgHead{$index} =~ s/($HAddrExp)/&link_refmsgid($1)/geo;
+	$Message{$index} =~ s/($HAddrExp)/&link_refmsgid($1)/geo;
 
 	print $msghandle "<!--X-Head-of-Message-->\n";
 	print $msghandle $MsgHead{$index};
@@ -1103,12 +1213,11 @@ sub output_mail {
     ($template = $MSGBODYEND) =~ s/$VarExp/&replace_li_var($1,$index)/geo;
     print $msghandle $template;
     if (!$nocustom && $DoFolRefs && defined($Follow{$index})) {
-	@array2 = split(/$bs/o, $Follow{$index});
-	if ($#array2 >= 0) {
+	if (scalar(@{$Follow{$index}})) {
 	    ($template = $FOLUPBEGIN) =~
 		s/$VarExp/&replace_li_var($1,$index)/geo;
 	    print $msghandle $template;
-	    foreach (@array2) {
+	    foreach (@{$Follow{$index}}) {
 		($template = $FOLUPLITXT) =~
 		    s/$VarExp/&replace_li_var($1,$_)/geo;
 		print $msghandle $template;
@@ -1126,12 +1235,12 @@ sub output_mail {
     }
     print $msghandle "<!--X-References-->\n";
     if (!$nocustom && $DoFolRefs && defined($Refs{$index})) {
-	@array2 = split(/$X/o, $Refs{$index});
 	$tmp2 = 0;	# flag for when first ref printed
-	if ($#array2 >= 0) {
-	    foreach (@array2) {
-		next  unless defined($MsgId{$_});
-		next  unless defined($IndexNum{$MsgId{$_}});
+	if (scalar(@{$Refs{$index}})) {
+	    my($ref_msgid, $ref_index, $ref_num);
+	    foreach $ref_msgid (@{$Refs{$index}}) {
+		next  unless defined($ref_index = $MsgId{$ref_msgid});
+		next  unless defined($ref_num = $IndexNum{$ref_index});
 		if (!$tmp2) {
 		    ($template = $REFSBEGIN) =~
 			s/$VarExp/&replace_li_var($1,$index)/geo;
@@ -1139,7 +1248,7 @@ sub output_mail {
 		    $tmp2 = 1;
 		}
 		($template = $REFSLITXT) =~
-		    s/$VarExp/&replace_li_var($1,$MsgId{$_})/geo;
+		    s/$VarExp/&replace_li_var($1,$ref_index)/geo;
 		print $msghandle $template;
 	    }
 
@@ -1186,41 +1295,51 @@ sub output_mail {
 	close($msginfh);
 	#&file_remove($tmppathname);
     }
-    &file_rename($tmppathname, $filepathname)  unless $SINGLE;
+    if (!$SINGLE) {
+	file_gzip($tmppathname)  if $GzipFiles;
+	file_chmod(file_rename($tmppathname, $filepathname));
+    }
 
     ## Create user defined files
+    my($drvfh);
     foreach (keys %UDerivedFile) {
 	($tmp = $_) =~ s/$VarExp/&replace_li_var($1,$index)/geo;
-	$tmp2 = join($DIRSEP, $OUTDIR, $tmp);
-	if ($drvfh = &file_create($tmp2, $GzipFiles)) {
-	    ($template = $UDerivedFile{$_}) =~
-		s/$VarExp/&replace_li_var($1,$index)/geo;
-	    print $drvfh $template;
-	    close($drvfh);
-	    if ($Derived{$index}) {
-		$Derived{$index} .= $X . $tmp;
-	    } else {
-		$Derived{$index} = $tmp;
-	    }
+	($drvfh, $tmppathname) = file_temp('drvXXXXXXXXXX', $OUTDIR);
+	($template = $UDerivedFile{$_}) =~
+	    s/$VarExp/&replace_li_var($1,$index)/geo;
+	print $drvfh $template;
+	close($drvfh);
+	file_gzip($tmppathname)  if $GzipFiles;
+	file_chmod(file_rename($tmppathname, join($DIRSEP, $OUTDIR, $tmp)));
+
+	if (defined($Derived{$index})) {
+	    push(@{$Derived{$index}}, $tmp);
 	} else {
-	    warn "Warning: Unable to create $tmp2\n";
+	    $Derived{$index} = [ $tmp ];
 	}
     }
-    if (defined($Derived{$index}) &&
-	    (@array2 = split(/$X/o, $Derived{$index}))) {
-	&remove_dups(*array2);
-	$Derived{$index} = join($X, @array2);
+    if (defined($Derived{$index})) {
+	$Derived{$index} = [ remove_dups($Derived{$index}) ];
     }
 
     ## Set modification times -- Use eval incase OS does not support utime.
     if ($MODTIME && !$SINGLE) {
 	eval {
-	    $tmp = &get_time_from_index($index);
-	    @array2 = split(/$X/o, $Derived{$index});
-	    grep($_ = $OUTDIR . $DIRSEP . $_, @array2);
+	    $tmp = get_time_from_index($index);
+	    if (defined($Derived{$index})) {
+	      @array2 = @{$Derived{$index}};
+	      grep($_ = $OUTDIR . $DIRSEP . $_, @array2);
+	    } else {
+	      @array2 = ( );
+	    }
 	    unshift(@array2, $filepathname);
-	    &file_utime($tmp, $tmp, @array2);
+	    file_utime($tmp, $tmp, @array2);
 	};
+	if ($@) {
+	    warn qq/\nWarning: Your platform does not support setting file/,
+		   qq/         modification times\n/;
+	    $MODTIME = 0;
+	}
     }
 
     ($i_p0, $filename);
@@ -1234,13 +1353,12 @@ sub output_mail {
 ##	delmsg delets a message from the archive.
 ##
 sub delmsg {
-    local($key) = @_;
-    local($filename);
-    local($pathname);
+    my $key = shift;
+    my($pathname);
 
     #&defineIndex2MsgId();
-    $msgnum = $IndexNum{$key};  return 0  if ($msgnum eq '');
-    $filename = join($DIRSEP, $OUTDIR, &msgnum_filename($msgnum));
+    my $msgnum = $IndexNum{$key};  return 0  if ($msgnum eq '');
+    my $filename = join($DIRSEP, $OUTDIR, &msgnum_filename($msgnum));
     delete $ContentType{$key};
     delete $Date{$key};
     delete $From{$key};
@@ -1248,15 +1366,14 @@ sub delmsg {
     delete $Refs{$key};
     delete $Subject{$key};
     delete $MsgId{$Index2MsgId{$key}};
-    &file_remove($filename)  unless $KeepOnRmm;
-    foreach $filename (split(/$X/o, $Derived{$key})) {
-	$pathname = (&OSis_absolute_path($filename)) ?
-			$filename :
-			join($DIRSEP, $OUTDIR, $filename);
+    file_remove($filename)  unless $KeepOnRmm;
+    foreach $filename (@{$Derived{$key}}) {
+	$pathname = (OSis_absolute_path($filename)) ?
+			$filename : join($DIRSEP, $OUTDIR, $filename);
 	if (-d $pathname) {
-	    &dir_remove($pathname)  unless $KeepOnRmm;
+	    dir_remove($pathname)  unless $KeepOnRmm;
 	} else {
-	    &file_remove($pathname)  unless $KeepOnRmm;
+	    file_remove($pathname)  unless $KeepOnRmm;
 	}
     }
     delete $Derived{$key};
@@ -1268,17 +1385,17 @@ sub delmsg {
 ##	Routine to convert a msgid to an anchor
 ##
 sub link_refmsgid {
-    local($refmsgid, $onlynew) = @_;
+    my $refmsgid = dehtmlize(shift);
+    my $onlynew  = shift;
 
     if (defined($MsgId{$refmsgid}) &&
-	defined($IndexNum{$MsgId{$refmsgid}}) &&
-	(!$onlynew || $NewMsgId{$refmsgid})) {
-	local($lreftmpl) = $MSGIDLINK;
+	    defined($IndexNum{$MsgId{$refmsgid}}) &&
+	    (!$onlynew || $NewMsgId{$refmsgid})) {
+	my($lreftmpl) = $MSGIDLINK;
 	$lreftmpl =~ s/$VarExp/&replace_li_var($1,$MsgId{$refmsgid})/geo;
-	$lreftmpl;
-    } else {
-	$refmsgid;
+	return $lreftmpl;
     }
+    htmlize($refmsgid);
 }
 
 ##---------------------------------------------------------------------------
